def get_cosine_similarity(vec_a: list[float], vec_b: list[float]):
    if len(vec_a) != len(vec_b):
        raise ValueError("Vector dimensions must match for calculating cosine distance")

    score = 0.0
    for i in range(0, len(vec_a)):
        score += vec_a[i] * vec_b[i]

    return score




CREATE_INDEX_MANAGER_TABLE = '''
CREATE TABLE IF NOT EXISTS index_manager (
    index_id BIGSERIAL PRIMARY KEY,
    name TEXT NOT NULL UNIQUE,
    dimensions INTEGER NOT NULL,
    max_elements INTEGER NOT NULL,
    m INTEGER NOT NULL,
    ef INTEGER NOT NULL,
    ef_construction INTEGER NOT NULL,
    allow_replace_deleted BOOLEAN NOT NULL
);
'''

CREATE_INDEX_DATA_TABLE = '''
CREATE TABLE IF NOT EXISTS index_data (
    vector_id BIGINT PRIMARY KEY GENERATED ALWAYS AS IDENTITY,
    id text NOT NULL,
    index_id BIGINT,
    text TEXT,
    embedding JSONB NOT NULL,
    metadata JSONB,
    UNIQUE(id, index_id),
    FOREIGN KEY(index_id) REFERENCES index_manager(index_id) ON DELETE CASCADE
);
'''

DELETE_VECTORS_FROM_INDEX = '''
DELETE FROM index_data
WHERE id = ANY(%s) AND index_id = %s
RETURNING vector_id
'''

GET_ALL_VECTORS = '''
SELECT vector_id, embedding, {}
FROM index_data
WHERE index_id = %s
'''

GET_ALL_INDEX_DETAILS = '''
SELECT index_id, name, dimensions, max_elements, m, ef, ef_construction, allow_replace_deleted
FROM index_manager
'''

GET_INDEX_DETAILS_BY_NAME = '''
SELECT index_id, name, dimensions, max_elements, m, ef, ef_construction, allow_replace_deleted
FROM index_manager
WHERE name = %s
'''

GET_VECTOR_IDS_OF_RESULTS = '''
SELECT vector_id, {}
FROM index_data
WHERE vector_id = ANY(%s) AND index_id = %s
ORDER BY vector_id ASC
'''

INSERT_DATA_TO_INDEX = '''
INSERT INTO index_data
(id, index_id, text, embedding, metadata)
VALUES(%s, %s, %s, %s, %s)
ON CONFLICT(id, index_id)
DO UPDATE SET id = %s, index_id = %s, text = %s, embedding = %s, metadata = %s
RETURNING vector_id
'''

INSERT_INDEX_TO_MANAGER = '''
INSERT INTO index_manager
(name, dimensions, max_elements, m, ef, ef_construction, allow_replace_deleted)
VALUES (%s, %s, %s, %s, %s, %s, %s);
'''

UPDATE_EF = '''
UPDATE index_manager
SET ef = %s
WHERE name = %s
'''



import os
import json
from typing import Dict, Optional, Tuple

def convert_row_to_dict(row: Tuple, include: Dict, with_embedding: bool = False):
    if with_embedding:
        # On Postgres embedding is stored as a list of floating points
        if isinstance(row[1], str):
            returning_dict = {"id": row[2], "embedding": json.loads(row[1])}
        else:
            returning_dict = {"id": row[2], "embedding": row[1]}

        if include["document"]:
            returning_dict["document"] = row[3]
            if include["metadata"]:
                returning_dict["metadata"] = row[4]
        elif include["metadata"]:
            returning_dict["metadata"] = row[3]

        return returning_dict
    else:
        returning_dict = {"id": row[1]}
        if include["document"]:
            returning_dict["document"] = row[2]
            if include["metadata"]:
                returning_dict["metadata"] = row[3]
        elif include["metadata"]:
            returning_dict["metadata"] = row[2]

        return returning_dict

def ensure_valid_path(persist_directory: str, file_name: Optional[str] = None) -> bool:
    """
    Creates required directories if they do not exist.

    If only persist_directory is passed, returns True.
    When file_name is passed, function returns boolean based on whether the
    file can be found in given path.
    """
    # Ensure directory exists
    if not (os.path.isdir(persist_directory)):
        os.makedirs(persist_directory)

    if file_name is None:
        return True

    if os.path.exists(
        os.path.join(persist_directory, file_name)
    ):
        return True
    else:
        return False



import numpy as np
from fastapi import FastAPI
from fastapi.responses import JSONResponse
from citrusdb import Client

app = FastAPI()
client = Client(persist_directory="citrus")

@app.post("/index")
def create_index(index_name: str, max_elements: int = 1000, allow_replace_deleted: bool = True, M = 64, ef_construction = 200):
    try:
        client.create_index(
            name=index_name,
            max_elements=int(max_elements),
            allow_replace_deleted=allow_replace_deleted,
            M=int(M),
            ef_construction=int(ef_construction)
        )
        return {"message": f"Index '{index_name}' created successfully."}
    except Exception as error:
        return JSONResponse(status_code=500, content=error)

@app.post("/insert")
def insert_elements(index_name: str, elements: list[dict]):
    try:
        embeddings = None
        texts = None
        ids = [element["id"] for element in elements]

        keys = elements[0].keys()
        if "embedding" in keys:
            embeddings = np.asarray([element["embedding"] for element in elements], np.float32)
        if "text" in keys:
            texts = [element["text"] for element in elements]

        metadatas = [element["metadata"] for element in elements]

        client.add(index=index_name, ids=ids, embeddings=embeddings, documents=texts, metadatas=metadatas)
        return {"message": f"{len(ids)} vectors indexed."}
    except Exception as error:
        return JSONResponse(status_code=500, content=error)

@app.get("/search")
def semantic_search(index_name: str, search: dict, k: int = 1):
    try:
        res = client.query(index=index_name, query_embeddings=np.asarray([search["embedding"]]), k=k)
        if res:
            results, distances = res
            response = []
            for i, result in enumerate(results):
                for j, id in result:
                    obj = {"id": id, "similarity_score": 1 - distances[i][j]}
                    response.append(obj)  # Assuming documents is accessible here
            return {"results": response}
    except Exception as error:
        return JSONResponse(status_code=500, content=error)

'''
if __name__ == "__main__":
    uvicorn.run(app, host="0.0.0.0", port=8000)
'''



from numpy import float32
from numpy._typing import NDArray
from typing import List, Optional, TypedDict

ID = str
IDs = List[ID]

class Document(TypedDict):
    id: int
    text: Optional[str]
    embedding: Optional[NDArray[float32]]



import sqlite3
from typing import Dict, List

class QueryBuilder:
    _con: sqlite3.Connection

    def __init__(self, connection):
        self._con = connection

    def build_query(self, index_name, filters):
        # Building the base SQL query
        sql_query = f"SELECT d.vector_id, d.id, d.index_id, d.text FROM index_data d JOIN index_manager m ON m.index_id = d.index_id WHERE m.name = '{index_name}'"

        # Adding the filter criteria to the SQL query
        if filters:
            sql_query += " AND "
            conditions = []
            for filter in filters:
                field = filter["field"]
                operator = filter["operator"]
                value = filter["value"]
                if isinstance(value, int) or isinstance(value, float):
                    condition = f"metadata ->> '{field}' {operator} {value}"
                else:
                    condition = f"metadata ->> '{field}' {operator} '{value}'"
                conditions.append(condition)
            sql_query += " AND ".join(conditions)

        return sql_query

    def execute_query(self, index_name: str, filters: List[Dict]):
        cursor = self._con.cursor()

        # Building and executing the query
        sql_query = self.build_query(index_name, filters)
        cursor.execute(sql_query)
        results = cursor.fetchall()

        return results




from abc import ABC, abstractmethod
from typing import Any, Dict, List, Tuple, Optional

from citrusdb.utils.types import IDs


class BaseDB(ABC):
    @abstractmethod
    def create_index(
        self,
        name: str,
        max_elements: int,
        M: int,
        ef_construction: int,
        allow_replace_deleted: bool,
        dimensions: Optional[int] = 1536,
    ):
        pass

    @abstractmethod
    def delete_vectors_from_index(
        self,
        index_id: int,
        ids: IDs
    ) -> List[int]:
        pass

    @abstractmethod
    def filter_vectors(self, index_name: str, filters: List[Dict]) -> List:
        pass

    @abstractmethod
    def get_indices(self) -> List[Any]:
        pass

    @abstractmethod
    def get_index_details(
        self,
        name: str
    ) -> Optional[Tuple[int, str, int, int, int, int, int, bool]]:
        pass

    @abstractmethod
    def get_all_vectors_in_index(
        self,
        name: str,
        include: Dict
    ) -> List[Dict]:
        pass

    @abstractmethod
    def get_vector_ids_of_results(
        self,
        name: str,
        results: List[List[int]],
        include: Dict
    ) -> List[List[Dict]]:
        pass

    @abstractmethod
    def insert_to_index(
        self,
        data
    ) -> List[int]:
        pass

    @abstractmethod
    def update_ef(
        self,
        name: str,
        ef: int
    ):
        pass




import os
import numpy as np
import openai
from typing import List

openai.api_key = os.getenv("OPENAI_API_KEY")


def get_embeddings(data: List[str], model="text-embedding-ada-002"):
    strings = [text.replace("\n", " ") for text in data]

    response = openai.Embedding.create(model=model, input=strings)

    return np.array([result["embedding"] for result in response["data"]], dtype=np.float32)



import os
import json
from typing import Any, Dict, List, Optional
from numpy import float32
from numpy._typing import NDArray
import shutil

from citrusdb.api.index import Index
from citrusdb.db import BaseDB
from citrusdb.db.postgres.db import PostgresDB
from citrusdb.db.sqlite.db import SQLiteDB
from citrusdb.embedding.utils import get_cosine_similarity
from citrusdb.utils.types import IDs


class LocalAPI:
    _db: Dict[str, Index] 
    _SQLClient: BaseDB
    persist_directory: Optional[str]
    _TEMP_DIRECTORY = "citrus_temp"

    def __init__(
        self,
        persist_directory: Optional[str] = None,
        database_type: Optional[str] = "sqlite",
        **kwargs: Optional[Dict[str, Any]]
    ):
        self._db = {}
        self.persist_directory = persist_directory

        if not(persist_directory) and os.path.isdir(self._TEMP_DIRECTORY):
            # Cleanup previous sqlite data
            shutil.rmtree(self._TEMP_DIRECTORY)

        if persist_directory and (database_type == "pg" or database_type == "postgres"):
            self._SQLClient = PostgresDB(**kwargs)
        else:
            self._SQLClient = SQLiteDB(persist_directory if persist_directory else self._TEMP_DIRECTORY)

    def create_index(
        self,
        name: str,
        max_elements: int = 1000,
        M: int = 64,
        ef_construction: int = 200,
        allow_replace_deleted: bool = False,
    ):
        if not(self._SQLClient.get_index_details(name)):
            self._SQLClient.create_index(
                name,
                max_elements,
                M,
                ef_construction,
                allow_replace_deleted
            )

        self._db[name] = Index(
            name=name,
            max_elements=max_elements,
            persist_directory=self.persist_directory,
            M=M,
            ef_construction=ef_construction,
            allow_replace_deleted=allow_replace_deleted
        )

    def add(
        self,
        index: str,
        ids: IDs,
        documents: Optional[List[str]] = None,
        embeddings: Optional[NDArray[float32]] = None,
        metadatas: Optional[List[Dict]] = None
    ):
        """
        Insert embeddings/text documents

        index: Name of index
        ids: Unique ID for each element
        documents: List of strings to index
        embeddings: List of embeddings to index
        metadatas: Additional metadata for each vector
        """

        if embeddings is None and documents is None:
            raise ValueError("Please provide either embeddings or documents.")

        index_details = self._SQLClient.get_index_details(index)
        if index_details is None:
            raise ValueError(f"Index with name '{index}' does not exist.")

        if (documents is not None) and (embeddings is None):
            from citrusdb.embedding.openai import get_embeddings

            embeddings = get_embeddings(documents)

        if embeddings is not None:
            embedding_dim = len(embeddings[0])
            index_id = index_details[0]
            index_dim = index_details[2]
            replace_deleted = True if index_details[7] else False

            # Check whether the dimensions are equal
            if embedding_dim != index_dim:
                raise ValueError(
                        f"Embedding dimension ({embedding_dim}) and index "
                        + f"dimension ({index_dim}) do not match."
                        )

            # Ensure no of ids = no of embeddings
            if len(ids) != len(embeddings):
                raise ValueError(f"Number of embeddings" + " and ids are different.")

            data = []
            for i in range(len(ids)):
                row = (
                    ids[i],
                    index_id,
                    None if documents is None else documents[i],
                    json.dumps(embeddings[i].tolist()),
                    None if metadatas is None else json.dumps(metadatas[i])
                )
                data.append(row + row)

            # Insert data into DB
            hnsw_labels = self._SQLClient.insert_to_index(data)

            # Index vectors
            self._db[index].add(
                ids=hnsw_labels,
                embeddings=embeddings,
                replace_deleted=replace_deleted
            )

    def delete_vectors(
        self,
        index: str,
        ids: IDs
    ):
        index_details = self._SQLClient.get_index_details(index)
        if index_details is None:
            raise ValueError(f"Could not find index: {index}")

        index_id = index_details[0]
        hnsw_labels = self._SQLClient.delete_vectors_from_index(
            index_id=index_id,
            ids=ids
        )

        self._db[index].delete_vectors(hnsw_labels)

    def reload_indices(self):
        """
        Load all indices from disk to memory
        """

        indices = self._SQLClient.get_indices()
        for index in indices:
            index_name = index[1]
            # Load index
            self.create_index(
                name=index_name,
                max_elements=index[3],
                M=index[4],
                ef_construction=index[6],
                allow_replace_deleted=index[7]
            )
            # Set ef value
            self._db[index_name].set_ef(index[5])

    def set_ef(self, index: str, ef: int):
        index_details = self._SQLClient.get_index_details(index)
        if index_details is None:
            raise ValueError(f"Could not find index: {index}")

        self._SQLClient.update_ef(index, ef)
        self._db[index].set_ef(ef)

    def query(
        self,
        index: str,
        documents: Optional[List[str]] = None,
        query_embeddings: Optional[NDArray[float32]] = None,
        k=1,
        filters: Optional[List[Dict]] = None,
        include: List[str] = []
    ):
        allowed_ids = []
        if filters is not None:
            allowed_ids = self._SQLClient.filter_vectors(index, filters)

        filter_function = lambda label: label in allowed_ids

        included_columns = {"id": True, "document": False, "metadata": False}
        if "document" in include:
            included_columns["document"] = True
        if "metadata" in include:
            included_columns["metadata"] = True

        flag = 1
        for key in self._db.keys():
            if key == index:
                flag = 0

                try:
                    results, distances = self._db[key].query(
                        documents=documents,
                        query_embeddings=query_embeddings,
                        k=k,
                        filter_function=None if filters is None else filter_function
                    )
                    elements = self._SQLClient.get_vector_ids_of_results(
                        name=index,
                        results=results,
                        include=included_columns
                    )
                    for i, rows in enumerate(elements):
                        for j, row in enumerate(rows):
                            row["distance"] = distances[i][j]
                    return elements
                except RuntimeError as error:
                    # Return all vectors currently in index
                    # Deals with error when top_k is higher than no. of vectors indexed.
                    # Error: Cannot return the results in a contigious 2D array. Probably ef or M is too small
                    print("ERROR:", error, "-- Falling back to existing vectors in index")
                    indexed_elements = self._SQLClient.get_all_vectors_in_index(
                        name=index,
                        include=included_columns
                    )
                    if (query_embeddings is None):
                        if documents is None:
                            raise Exception("Neither documents nor query embeddings provided.")
                        from citrusdb.embedding.openai import get_embeddings

                        embeddings = get_embeddings(documents)
                        query_embeddings = embeddings

                    returning_elements = []
                    for i, query in enumerate(query_embeddings):
                        results = []
                        for element in indexed_elements:
                            item = {**element}
                            item["distance"] = 1 - get_cosine_similarity(query, element["embedding"])
                            del item["embedding"]
                            results.append(item)
                        results.sort(key=lambda e : e["distance"])
                        returning_elements.append(results)
                    return returning_elements
        if flag:
            raise ValueError(f"Could not find index: {index}")

    def get_status(self, index: str):
        flag = 1
        for key in self._db.keys():
            if key == index:
                flag = 0
                self._db[key].get_status()

        if flag:
            raise ValueError(f"Could not find index: {index}")



import os
import sqlite3
from typing import Dict, List, Optional, Tuple

from citrusdb.db import BaseDB
from citrusdb.utils.types import IDs
from citrusdb.utils.utils import convert_row_to_dict, ensure_valid_path
import citrusdb.db.sqlite.queries as queries
from citrusdb.db.sqlite.query_builder import QueryBuilder


class SQLiteDB(BaseDB):
    _con: sqlite3.Connection

    def __init__(
        self,
        persist_directory: str,
    ):
        ensure_valid_path(persist_directory)

        self._con = sqlite3.connect(
            os.path.join(
                persist_directory, "citrus.db"
            )
        )

        cur = self._con.cursor()
        cur.execute("PRAGMA foreign_keys = ON")    # Enable foreign keys
        cur.executescript(f'''
        BEGIN;
        {queries.CREATE_INDEX_MANAGER_TABLE}
        {queries.CREATE_INDEX_DATA_TABLE}
        END;
        ''')
        cur.close()

    def create_index(
        self,
        name: str,
        max_elements: int,
        M: int,
        ef_construction: int,
        allow_replace_deleted: bool,
        dimensions: Optional[int] = 1536,
    ):
        cur = self._con.cursor()
        ef = ef_construction
        parameters = (name, dimensions, max_elements, M, ef, ef_construction, allow_replace_deleted)
        cur.execute(queries.INSERT_INDEX_TO_MANAGER, parameters)
        self._con.commit()
        cur.close()

    def delete_vectors_from_index(
        self,
        index_id: int,
        ids: IDs
    ):
        cur = self._con.cursor()
        query = queries.DELETE_VECTORS_FROM_INDEX.format(", ".join("?" * len(ids)))
        parameters = tuple(ids) + (index_id,)
        cur.execute(query, parameters)

        rows = cur.fetchall()
        self._con.commit()
        cur.close()

        vector_ids = [row[0] for row in rows]
        return vector_ids

    def filter_vectors(self, index_name: str, filters: List[Dict]):
        query_builder = QueryBuilder(self._con)
        res = query_builder.execute_query(index_name, filters)
        allowed_ids = []
        for row in res:
            allowed_ids.append(row[0])
        return allowed_ids

    def get_indices(self):
        """
        Fetch all index details from index_manager table.
        Returns a list of tuples where each one corresponds to an index.
        """

        cur = self._con.cursor()
        res = cur.execute(queries.GET_ALL_INDEX_DETAILS)
        rows = res.fetchall()
        cur.close()
        return rows

    def get_index_details(
        self,
        name: str
    ) -> Optional[Tuple[int, str, int, int, int, int, int, bool]]:
        cur = self._con.cursor()
        parameters = (name,)
        res = cur.execute(queries.GET_INDEX_DETAILS_BY_NAME, parameters)
        row = res.fetchone()
        cur.close()
        return row

    def get_all_vectors_in_index(self, name: str, include: Dict) -> List[Dict]:
        cols = "id"
        if include["document"]:
            cols += ", text"
            if include["metadata"]:
                cols += ", metadata"
        elif include["metadata"]:
            cols += ", metadata"

        index_details = self.get_index_details(name)
        index_id = index_details[0]                 # type: ignore

        cur = self._con.cursor()
        query = queries.GET_ALL_VECTORS.format(cols)
        parameters = (index_id,)
        res = cur.execute(query, parameters)
        rows = res.fetchall()
        cur.close()

        return [convert_row_to_dict(row=row, include=include, with_embedding=True) for row in rows]

    def get_vector_ids_of_results(
        self,
        name: str,
        results: List[List[int]],
        include: Dict
    ) -> List[List[Dict]]:
        cols = "id"
        if include["document"]:
            cols += ", text"
            if include["metadata"]:
                cols += ", metadata"
        elif include["metadata"]:
            cols += ", metadata"

        returning_list = []
        index_details = self.get_index_details(name)
        index_id = index_details[0]                 # type: ignore

        cur = self._con.cursor()
        for ids in results:
            query = queries.GET_VECTOR_IDS_OF_RESULTS.format(cols, ", ".join("?" * len(ids)))
            parameters = ()
            for id in ids:
                parameters += (int(id),)
            parameters += (index_id,)
            res = cur.execute(query, parameters)
            unordered_rows = res.fetchall()         # Rows not ordered according to similarity score

            # Order rows according to order of id in ids list
            ordered_rows = []
            for id in ids:
                low = 0; high = len(unordered_rows) - 1
                while (low <= high):
                    mid = low + (high - low)//2
                    curr_vector_id = unordered_rows[mid][0]
                    if curr_vector_id == id:
                        ordered_rows.append(
                            convert_row_to_dict(
                                row=unordered_rows[mid],
                                include=include
                            )
                        )
                        break
                    elif curr_vector_id < id:
                        low = mid + 1
                    else:
                        high = mid - 1

            returning_list.append(ordered_rows)
        cur.close()

        return returning_list

    def insert_to_index(
        self,
        data
    ):
        cur = self._con.cursor()
        vector_ids = []
        for row in data:
            res = cur.execute(queries.INSERT_DATA_TO_INDEX, row)
            vector_ids.append(res.fetchone()[0])

        self._con.commit()
        cur.close()

        return vector_ids

    def update_ef(
        self,
        name: str,
        ef: int
    ):
        cur = self._con.cursor()
        parameters = (ef, name)
        cur.execute(queries.UPDATE_EF, parameters)
        self._con.commit()
        cur.close()



from psycopg_pool import ConnectionPool
from psycopg import sql
from typing import Any, Dict, List, Optional, Tuple
from citrusdb.db import BaseDB
import citrusdb.db.postgres.queries as queries
from citrusdb.db.postgres.query_builder import QueryBuilder
from citrusdb.utils.types import IDs
from citrusdb.utils.utils import convert_row_to_dict


class PostgresDB(BaseDB):
    _pool: ConnectionPool

    def __init__(
        self,
        **kwargs: Dict[str, Any]
    ):
        # Setup connection pool
        self._pool = ConnectionPool(kwargs=kwargs)

        # Create index_manager and index_data table if they don't exist already
        with self._pool.connection() as conn:
            with conn.cursor() as cur:
                cur.execute(queries.CREATE_INDEX_MANAGER_TABLE)
                cur.execute(queries.CREATE_INDEX_DATA_TABLE)
                conn.commit()

    def create_index(
        self,
        name: str,
        max_elements: int,
        M: int,
        ef_construction: int,
        allow_replace_deleted: bool,
        dimensions: Optional[int] = 1536,
    ):
        ef = ef_construction
        parameters = (name, dimensions, max_elements, M, ef, ef_construction, allow_replace_deleted)
        # Create new index entry to postgres db
        with self._pool.connection() as conn:
            with conn.cursor() as cur:
                cur.execute(queries.INSERT_INDEX_TO_MANAGER, parameters)
                conn.commit()

    def delete_vectors_from_index(
        self,
        index_id: int,
        ids: IDs
    ):
        """
        Delete vectors with given list of IDs from specific index

        index_id: ID of index where the elements belong
        ids: List of IDs to be deleted
        """

        vector_ids = []
        parameters = [ids, index_id]
        with self._pool.connection() as conn:
            with conn.cursor() as cur:
                for vector_id in cur.execute(queries.DELETE_VECTORS_FROM_INDEX, parameters):
                    vector_ids.append(vector_id[0])
                conn.commit()

        return vector_ids

    def filter_vectors(self, index_name: str, filters: List[Dict]):
        """
        Get list of IDs of vectors that match filters.

        index_name: Name of index where the elements belong
        filters: List of filters to be applied
        """

        with self._pool.connection() as conn:
            query_builder = QueryBuilder(conn)
            res = query_builder.execute_query(index_name, filters)
            allowed_ids = []
            for row in res:
                allowed_ids.append(row[0])
            return allowed_ids

    def get_indices(self):
        """
        Get all index details
        """
        with self._pool.connection() as conn:
            with conn.cursor() as cur:
                cur.execute(queries.GET_ALL_INDEX_DETAILS)
                return cur.fetchall()

    def get_index_details(
        self,
        name: str
    ) -> Optional[Tuple[int, str, int, int, int, int, int, bool]]:
        """
        Get specific index details

        name: Name of index to fetch
        """
        parameters = (name,)
        with self._pool.connection() as conn:
            with conn.cursor() as cur:
                cur.execute(queries.GET_INDEX_DETAILS_BY_NAME, parameters)
                return cur.fetchone()

    def get_all_vectors_in_index(self, name: str, include: Dict) -> List[Dict]:
        """
        Get all vectors in index

        name: Name of index
        include: Dictionary of columns to be returned
        """
        index_details = self.get_index_details(name)
        if index_details is None:
            raise ValueError(f"Index '{name}' does not exist")

        cols = [sql.Identifier("id")]
        if include["document"]:
            cols.append(sql.Identifier("text"))
            if include["metadata"]:
                cols.append(sql.Identifier("metadata"))
        elif include["metadata"]:
            cols.append(sql.Identifier("metadata"))

        index_id = index_details[0]
        parameters = (index_id,)
        with self._pool.connection() as conn:
            with conn.cursor() as cur:
                query = sql.SQL(queries.GET_ALL_VECTORS).format(
                    sql.SQL(", ").join(cols)
                )
                cur.execute(query, parameters)
                rows = cur.fetchall()
                return [convert_row_to_dict(row=row, include=include, with_embedding=True) for row in rows]

    def get_vector_ids_of_results(
        self,
        name: str,
        results: List[List[int]],
        include: Dict
    ):
        """
        Get user facing IDs of results

        name: Name of index
        results: List of list of integer HNSW labels
        include: Dictionary of columns to be returned
        """
        index_details = self.get_index_details(name)
        if index_details is None:
            raise ValueError(f"Index '{name}' does not exist")

        cols = [sql.Identifier("id")]
        if include["document"]:
            cols.append(sql.Identifier("text"))
            if include["metadata"]:
                cols.append(sql.Identifier("metadata"))
        elif include["metadata"]:
            cols.append(sql.Identifier("metadata"))

        returning_list = []
        unordered_rows_list = []
        index_id = index_details[0]

        with self._pool.connection() as conn:
            with conn.cursor() as cur:
                data = []
                for ids in results:
                    ids_list = []
                    for id in ids:
                        ids_list.append(int(id))
                    data.append([ids_list, index_id])

                query = sql.SQL(queries.GET_VECTOR_IDS_OF_RESULTS).format(
                    sql.SQL(", ").join(cols)
                )
                cur.executemany(
                    query,
                    data,
                    returning=True
                )
                while True:
                    rows = cur.fetchall()
                    unordered_rows_list.append(rows)
                    if not cur.nextset():
                        break;

                conn.commit()

        # Order rows according to order of id in ids list
        for i, ids in enumerate(results):
            unordered_rows = unordered_rows_list[i]
            ordered_rows = []
            for id in ids:
                low = 0; high = len(unordered_rows) - 1
                while (low <= high):
                    mid = low + (high - low)//2
                    curr_vector_id = unordered_rows[mid][0]
                    if curr_vector_id == id:
                        ordered_rows.append(
                            convert_row_to_dict(
                                row=unordered_rows[mid],
                                include=include
                            )
                        )
                        break
                    elif curr_vector_id < id:
                        low = mid + 1
                    else:
                        high = mid - 1

            returning_list.append(ordered_rows)
        return returning_list

    def insert_to_index(
        self,
        data
    ):
        """
        Insert vectors to index

        data: Tuple of tuples corresponding to each row
        """
        vector_ids = []
        with self._pool.connection() as conn:
            with conn.cursor() as cur:
                cur.executemany(queries.INSERT_DATA_TO_INDEX, data, returning=True)
                while True:
                    vector_ids.append(cur.fetchone()[0])        # type: ignore
                    if not cur.nextset():
                        break;

                conn.commit()

        return vector_ids

    def update_ef(
        self,
        name: str,
        ef: int
    ):
        """
        Update ef for an index

        name: Name of index to be updated
        ef: New ef value
        """
        parameters = (ef, name)
        with self._pool.connection() as conn:
            with conn.cursor() as cur:
                cur.execute(queries.UPDATE_EF, parameters)
                conn.commit()



import hnswlib


class HnswIndex:
    _id: str
    _index: hnswlib.Index

    def __init__(self, id, space="cosine", dim=1536):
        self._id = id
        self._index = hnswlib.Index(space=space, dim=dim)

    def init_index(
        self, max_elements=1000, M=64, ef_construction=200, allow_replace_deleted=False
    ):
        self._index.init_index(max_elements, M, ef_construction, allow_replace_deleted)

    def add_items(self, data, ids, replace_deleted=False):
        max_elements = self._index.max_elements
        curr_elements = self._index.element_count

        # Increase index size
        if curr_elements + len(data) > max_elements:
            new_size = max(curr_elements + len(data), 2 * max_elements)
            self._index.resize_index(new_size)

        self._index.add_items(data, ids, replace_deleted)

    def knn_query(self, query_embeddings, k=1, filter_function=None):
        labels, distances = self._index.knn_query(query_embeddings, k, filter=filter_function)
        return labels, distances

    def set_ef(self, ef: int):
        self._index.set_ef(ef)

    def get_dimension(self):
        return self._index.dim

    def load_index(self, path: str, allow_replace_deleted=False):
        self._index.load_index(path, allow_replace_deleted=allow_replace_deleted)

    def mark_deleted(self, label: int):
        self._index.mark_deleted(label)

    def save_index(self, path: str):
        self._index.save_index(path)

    def get_status(self):
        print("Max elements", self._index.max_elements)
        print("Current elements", self._index.element_count)




from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="citrusdb",
    version="0.5.4",
    author="Debabrata Mondal",
    author_email="debabrata.js@protonmail.com",
    description="(distributed) vector database",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/0xDebabrata/citrus",
    packages=(
        find_packages(
            exclude=["demo"]
        ) +
        ["citrusdb.db.index", "citrusdb.db.sqlite", "citrusdb.db.postgres"]
    ),
    include_package_data=True,
    install_requires=[
        "numpy",
        "hnswlib",
        "openai",
        "psycopg[c]",
        "psycopg[pool]"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)



from typing import Any, Dict, Optional

def Client(
    persist_directory: Optional[str] = None,
    database_type: Optional[str] = "sqlite",
    **kwargs
):
    from citrusdb.api.local import LocalAPI

    return LocalAPI(persist_directory, database_type=database_type, **kwargs)



CREATE_INDEX_MANAGER_TABLE = '''
CREATE TABLE IF NOT EXISTS index_manager (
    index_id INTEGER PRIMARY KEY AUTOINCREMENT,
    name TEXT NOT NULL UNIQUE,
    dimensions INTEGER NOT NULL,
    max_elements INTEGER NOT NULL,
    m INTEGER NOT NULL,
    ef INTEGER NOT NULL,
    ef_construction INTEGER NOT NULL,
    allow_replace_deleted INTEGER NOT NULL
);
'''

CREATE_INDEX_DATA_TABLE = '''
CREATE TABLE IF NOT EXISTS index_data (
    vector_id INTEGER PRIMARY KEY AUTOINCREMENT,
    id TEXT NOT NULL,
    index_id INTEGER,
    text TEXT,
    embedding BLOB NOT NULL,
    metadata TEXT,
    UNIQUE(id, index_id),
    FOREIGN KEY(index_id) REFERENCES index_manager(index_id) ON DELETE CASCADE
);
'''

DELETE_VECTORS_FROM_INDEX = '''
DELETE FROM index_data
WHERE id IN ({}) AND index_id = ?
RETURNING vector_id
'''

GET_ALL_INDEX_DETAILS = '''
SELECT index_id, name, dimensions, max_elements, m, ef, ef_construction, allow_replace_deleted
FROM index_manager
'''

GET_ALL_VECTORS = '''
SELECT vector_id, embedding, {}
FROM index_data
WHERE index_id = ?
'''

GET_INDEX_DETAILS_BY_NAME = '''
SELECT index_id, name, dimensions, max_elements, m, ef, ef_construction, allow_replace_deleted
FROM index_manager
WHERE name = ?
'''

GET_VECTOR_IDS_OF_RESULTS = '''
SELECT vector_id, {}
FROM index_data
WHERE vector_id IN ({}) AND index_id = ?
ORDER BY vector_id ASC
'''

INSERT_DATA_TO_INDEX = '''
INSERT INTO index_data
(id, index_id, text, embedding, metadata)
VALUES(?, ?, ?, ?, ?)
ON CONFLICT(id, index_id)
DO UPDATE SET id = ?, index_id = ?, text = ?, embedding = ?, metadata = ?
RETURNING vector_id
'''

INSERT_INDEX_TO_MANAGER = '''
INSERT INTO index_manager
(name, dimensions, max_elements, m, ef, ef_construction, allow_replace_deleted)
VALUES (?, ?, ?, ?, ?, ?, ?);
'''

UPDATE_EF = '''
UPDATE index_manager
SET ef = ?
WHERE name = ?
'''



import os
import pickle
from typing import Any, List, Optional
from numpy import float32
from numpy._typing import NDArray
from citrusdb.db.index.hnswlib import HnswIndex
from citrusdb.utils.types import IDs
from citrusdb.utils.utils import ensure_valid_path


class Index:
    _db: HnswIndex
    _parameters: dict

    def __init__(
        self,
        name: str,
        max_elements: int = 1000,
        persist_directory: Optional[str] = None,
        M: int = 64,
        ef_construction: int = 200,
        allow_replace_deleted: bool = False,
    ):
        self._db = HnswIndex(id=name)

        self._parameters = {
            "index_name": name,
            "max_elements": max_elements,
            "persist_directory": persist_directory,
            "M": M,
            "ef_construction": ef_construction,
            "allow_replace_deleted": allow_replace_deleted,
        }

        if persist_directory:
            self._load_params()

            if ensure_valid_path(persist_directory, str(self._parameters["index_name"])):
                self._db.load_index(
                    os.path.join(
                        persist_directory, str(self._parameters["index_name"])
                    ),
                    allow_replace_deleted=bool(
                        self._parameters["allow_replace_deleted"]
                    ),
                )
            else:
                self._db.init_index(
                    max_elements=max_elements,
                    M=M,
                    ef_construction=ef_construction,
                    allow_replace_deleted=allow_replace_deleted,
                )
                self._save()
        else:
            self._db.init_index(
                max_elements=max_elements,
                M=M,
                ef_construction=ef_construction,
                allow_replace_deleted=allow_replace_deleted,
            )

    def add(
        self,
        ids: List[int],
        embeddings: Optional[NDArray[float32]],
        replace_deleted: bool,
    ):
        self._db.add_items(embeddings, ids, replace_deleted)
        if self._parameters["persist_directory"]:
            self._save()

    def _load_params(self):
        if ensure_valid_path(self._parameters["persist_directory"], ".citrus_params"):
            filename = os.path.join(
                self._parameters["persist_directory"], ".citrus_params"
            )
            with open(filename, "rb") as f:
                self._parameters = pickle.load(f)

    def delete_vectors(self, ids: List[int]):
        for id in ids:
            self._db.mark_deleted(id)
        if self._parameters["persist_directory"]:
            self._save()

    def _save(self):
        self._db.save_index(
            os.path.join(
                self._parameters["persist_directory"], self._parameters["index_name"]
            )
        )
        self._save_params()

    def _save_params(self):
        output_file = os.path.join(
            self._parameters["persist_directory"], ".citrus_params"
        )
        with open(output_file, "wb") as f:
            pickle.dump(self._parameters, f)

    def set_ef(self, ef: int):
        self._db.set_ef(ef)

    def query(
        self,
        documents: Optional[List[str]] = None,
        query_embeddings: Optional[NDArray[float32]] = None,
        k=1,
        filter_function=None
    ):
        if query_embeddings is None and documents is None:
            raise ValueError("Please provide either an embedding" + " or a document.")

        if (query_embeddings is None) and (documents is not None):
            from citrusdb.embedding.openai import get_embeddings

            embeddings = get_embeddings(documents)
            query_embeddings = embeddings

        return self._db.knn_query(query_embeddings, k, filter_function)

    def get_status(self):
        self._db.get_status()

    def get_dimension(self):
        return self._db.get_dimension()

    def get_replace_deleted(self):
        return self._parameters["allow_replace_deleted"]

